from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import shutil
import tempfile
import tensorflow as tf

from ..export import export_hub_module
from ..column import text_embedding_column, sequence_text_embedding_column


class TestTextEmbeddingColumn(tf.test.TestCase):
    def setUp(self):
        self.temp_dir = tempfile.mkdtemp()
        self.temp_file = os.path.join(self.temp_dir, 'source.txt')

        source = '\n'.join([
            'key1 1. 2. 3.',
            'key2 4. 5. 6.',
        ]) + '\n'
        with open(self.temp_file, 'wb') as ttf:
            ttf.write(source.encode('utf-8'))

        export_hub_module(self.temp_file, self.temp_dir)

        tf.reset_default_graph()

    def tearDown(self):
        shutil.rmtree(self.temp_dir, ignore_errors=True)

    def testContextDense1D(self):
        expected_value = [
            [1., 2., 3.],
            [0., 0., 0.],
        ]

        features = {
            'tokens': ['key1', 'key999']
        }
        context_input = tf.feature_column.input_layer(
            features=features,
            feature_columns=[
                text_embedding_column(key='tokens', module_spec=self.temp_dir)
            ]
        )

        with self.test_session() as sess:
            sess.run([tf.global_variables_initializer(), tf.tables_initializer()])
            features_value = sess.run(context_input)

        self.assertListEqual(expected_value, features_value.tolist())

    def testContextSparse1D(self):
        expected_value = [
            [1., 2., 3.],
            [0., 0., 0.],
        ]

        features = {
            'tokens': tf.SparseTensor(
                indices=[[0], [1]],
                values=['key1', 'key999'],
                dense_shape=[2]
            )
        }
        context_input = tf.feature_column.input_layer(
            features=features,
            feature_columns=[
                text_embedding_column(key='tokens', module_spec=self.temp_dir)
            ]
        )

        with self.test_session() as sess:
            sess.run([tf.global_variables_initializer(), tf.tables_initializer()])
            features_value = sess.run(context_input)

        self.assertListEqual(expected_value, features_value.tolist())

    def testContextSparse2D(self):
        expected_value = [
            [2.5, 3.5, 4.5],
            [0., 0., 0.],
        ]

        features = {
            'tokens': tf.SparseTensor(
                indices=[[0, 0], [0, 1], [1, 0]],
                values=['key1', 'key2', 'key999'],
                dense_shape=[2, 2]
            )
        }
        context_input = tf.feature_column.input_layer(
            features=features,
            feature_columns=[
                text_embedding_column(key='tokens', module_spec=self.temp_dir)
            ]
        )

        with self.test_session() as sess:
            sess.run([tf.global_variables_initializer(), tf.tables_initializer()])
            features_value = sess.run(context_input)

        self.assertListEqual(expected_value, features_value.tolist())

    def testSequenceSparse2D(self):
        expected_value = [
            [[1., 2., 3.], [4., 5., 6.], [0., 0., 0.]],
            [[4., 5., 6.], [0., 0., 0.], [0., 0., 0.]],
        ]

        features = {
            'tokens': tf.SparseTensor(
                indices=[[0, 0], [0, 1], [0, 2], [1, 0]],
                values=['key1', 'key2', 'key999', 'key2'],
                dense_shape=[2, 3],
            ),
        }
        sequence_input = tf.contrib.feature_column.sequence_input_layer(
            features=features,
            feature_columns=[
                sequence_text_embedding_column(key='tokens', module_spec=self.temp_dir)
            ])

        with self.test_session() as sess:
            sess.run([tf.global_variables_initializer(), tf.tables_initializer()])
            features_value, features_length = sess.run(sequence_input)

        self.assertListEqual(expected_value, features_value.tolist())
        self.assertListEqual([3, 1], features_length.tolist())

    def testSequenceSparse3D(self):
        expected_value = [
            [[2.5, 3.5, 4.5], [0., 0., 0.]],
            [[4., 5., 6.], [0., 0., 0.]],
        ]

        features = {
            'tokens': tf.SparseTensor(
                indices=[[0, 0, 0], [0, 0, 1], [0, 1, 0], [1, 0, 0]],
                values=['key1', 'key2', 'key999', 'key2'],
                dense_shape=[2, 2, 2],
            ),
        }

        sequence_input = tf.contrib.feature_column.sequence_input_layer(
            features=features,
            feature_columns=[
                sequence_text_embedding_column(key='tokens', module_spec=self.temp_dir)
            ])

        with self.test_session() as sess:
            sess.run([tf.global_variables_initializer(), tf.tables_initializer()])
            features_value, features_length = sess.run(sequence_input)

        self.assertListEqual(expected_value, features_value.tolist())
        self.assertListEqual([2, 1], features_length.tolist())



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import tensorflow as tf
from tensorflow.python.framework import dtypes
from tensorflow.python.feature_column import feature_column
from tensorflow_hub import module


def text_embedding_column(key, module_spec, trainable=False):
    """Upgraded version of tensorflow_hub.text_embedding_column

    Returns:
      `_SequenceDenseColumn` that converts from text input.

    Raises:
       ValueError: if module_spec is not suitable for use in this feature column.
    """

    module_spec = module.as_module_spec(module_spec)
    _check_module_is_text_embedding(module_spec)

    return _TextEmbeddingColumn(
        key=key,
        module_spec=module_spec,
        trainable=trainable,
        signature='context',
    )


def sequence_text_embedding_column(key, module_spec, trainable=False):
    """Upgraded version of tensorflow_hub.text_embedding_column with sequential input support

    Returns:
      `_SequenceDenseColumn` that converts from text input.

    Raises:
       ValueError: if module_spec is not suitable for use in this feature column.
    """

    module_spec = module.as_module_spec(module_spec)
    _check_module_is_sequence_text_embedding(module_spec)

    return _TextEmbeddingColumn(
        key=key,
        module_spec=module_spec,
        trainable=trainable,
        signature='sequence',
    )


class _TextEmbeddingColumn(
    feature_column._DenseColumn,
    feature_column._SequenceDenseColumn,
    collections.namedtuple('_ModuleEmbeddingColumn', ('key', 'module_spec', 'trainable', 'signature'))):
    """Returned by text_embedding_column() or sequence_text_embedding_column(). Do not use directly."""

    @property
    def name(self):
        if not hasattr(self, '_name'):
            self._name = '{}_{}_hub_text_embedding'.format(self.key, self.signature)
        return self._name

    @property
    def _parse_example_spec(self):
        return {self.key: tf.VarLenFeature(tf.string)}

    def _transform_feature(self, inputs):
        input_tensor = inputs.get(self.key)
        input_tensor = feature_column._to_sparse_input_and_drop_ignore_values(input_tensor)

        if input_tensor.dtype != dtypes.string:
            raise ValueError('column {} dtype must be string, got: {}.'.format(self.key, input_tensor.dtype))

        return input_tensor

    @property
    def _variable_shape(self):
        if not hasattr(self, '_shape'):
            output_info = self.module_spec.get_output_info_dict(signature=self.signature)
            embed_size = output_info['default'].get_shape()[-1]
            self._shape = tf.TensorShape([embed_size])
        return self._shape

    @property
    def _hub_module(self):
        if not hasattr(self, '_module'):
            self._module = module.Module(self.module_spec, trainable=self.trainable)
        return self._module

    def _get_dense_tensor(self, inputs, weight_collections=None, trainable=None):
        del weight_collections, trainable

        if self.signature != 'context':
            raise ValueError(
                'Column {} could not be used as context feature column. '
                'Use text_embedding_column instead'.format(self.name)
            )

        sparse_keys = inputs.get(self)
        sparse_keys.shape.with_rank_at_least(1)
        sparse_keys.shape.with_rank_at_most(2)

        batch_size = sparse_keys.dense_shape[0]
        sparse_keys = tf.sparse_reshape(sparse_keys, shape=tf.stack([batch_size, -1]))

        return self._hub_module(sparse_keys, signature=self.signature)

    def _get_sequence_dense_tensor(self, inputs, weight_collections=None, trainable=None):
        del weight_collections, trainable

        if self.signature != 'sequence':
            raise ValueError(
                'Column {} could not be used as sequence feature column. '
                'Use sequence_text_embedding_column instead'.format(self.name)
            )

        sparse_keys = inputs.get(self)
        sparse_keys.shape.with_rank_at_least(2)
        sparse_keys.shape.with_rank_at_most(3)

        batch_size, max_length = sparse_keys.dense_shape[0], sparse_keys.dense_shape[1]
        sparse_keys = tf.sparse_reshape(sparse_keys, shape=tf.stack([batch_size, max_length, -1]))

        dense_tensor = self._hub_module(sparse_keys, signature=self.signature)
        sequence_length = feature_column._sequence_length_from_sparse_tensor(inputs.get(self))

        return feature_column._SequenceDenseColumn.TensorSequenceLengthPair(
            dense_tensor=dense_tensor, sequence_length=sequence_length)


def _check_module_is_text_embedding(module_spec):
    """Raises ValueError if `module_spec` is not a text-embedding module.

    Args:
      module_spec: A `ModuleSpec` to test.

    Raises:
      ValueError: if `module_spec` default signature is not compatible with
      SparseTensor(string, shape=(?,)) -> Tensor(float32, shape=(?,K)).
    """
    issues = []

    # Find issues with signature inputs.
    input_info_dict = module_spec.get_input_info_dict(signature='context')
    if len(input_info_dict) != 1:
        issues.append('Module "context" signature must require only one input')
    else:
        input_info, = input_info_dict.values()
        input_shape = input_info.get_shape()
        if not (input_info.dtype == tf.string and
                input_info.is_sparse and
                input_shape.ndims == 2 and
                input_shape.as_list() == [None, None]):
            issues.append(
                'Module "context" signature must have only one input '
                'tf.SparseTensor(shape=(?,?), dtype=string)'
            )

    # Find issues with signature outputs.
    output_info_dict = module_spec.get_output_info_dict(signature='context')
    if 'default' not in output_info_dict:
        issues.append('Module "context" signature must have a "default" output.')
    else:
        output_info = output_info_dict['default']
        output_shape = output_info.get_shape()
        if not (output_info.dtype == tf.float32 and
                output_shape.ndims == 2 and
                not output_shape.as_list()[0] and
                output_shape.as_list()[1]):
            issues.append(
                'Module "context" signature must have only one output '
                'tf.Tensor(shape=(?,K), dtype=float32).'
            )

    if issues:
        raise ValueError('Module is not a context text-embedding: %r' % issues)

def _check_module_is_sequence_text_embedding(module_spec):
    """Raises ValueError if `module_spec` is not a text-embedding module.

    Args:
      module_spec: A `ModuleSpec` to test.

    Raises:
      ValueError: if `module_spec` default signature is not compatible with
      SparseTensor(string, shape=(?,)) -> Tensor(float32, shape=(?,K)).
    """
    issues = []

    # Find issues with signature inputs.
    input_info_dict = module_spec.get_input_info_dict(signature='sequence')
    if len(input_info_dict) != 1:
        issues.append('Module "sequence" signature must require only one input')
    else:
        input_info, = input_info_dict.values()
        input_shape = input_info.get_shape()
        if not (input_info.dtype == tf.string and
                input_info.is_sparse and
                input_shape.ndims == 3 and
                input_shape.as_list() == [None, None, None]):
            issues.append(
                'Module "sequence" signature must have only one input '
                'tf.SparseTensor(shape=(?,?,?), dtype=string)'
            )

    # Find issues with signature outputs.
    output_info_dict = module_spec.get_output_info_dict(signature='sequence')
    if 'default' not in output_info_dict:
        issues.append('Module "sequence" signature must have a "default" output.')
    else:
        output_info = output_info_dict['default']
        output_shape = output_info.get_shape()
        if not (output_info.dtype == tf.float32 and
                output_shape.ndims == 3 and
                not output_shape.as_list()[0] and
                not output_shape.as_list()[1] and
                output_shape.as_list()[2]):
            issues.append(
                'Module "sequence" signature must have only one output '
                'tf.Tensor(shape=(?,?,K), dtype=float32).'
            )

    if issues:
        raise ValueError('Module is not a sequence text-embedding: %r' % issues)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import argparse
import collections
import logging
import numpy as np
import os
import tensorflow as tf
import tensorflow_hub as hub

_EMBEDDINGS_VAR_NAME = 'embeddings'
_UNIQUE_KEY_NAME = '<UNQ>'


def _iterate_text_vectors(file_name):
    with open(file_name, 'rb') as src_file:
        for row in src_file:
            row = row.decode('utf-8').strip().split()
            if len(row) < 2:  # at least label and one feature
                continue
            yield row


def read_and_count(file_name):
    key_values = []
    unique_features = None
    features_size = None

    for row in _iterate_text_vectors(file_name):
        if features_size is None:
            features_size = len(row) - 1
        else:
            message = 'All embeddings should have same size (inferred {}), but row {} has size {}'
            assert features_size == len(row) - 1, message.format(features_size, row[0], len(row) - 1)

        key_values.append(row[0])
        if row[0] == _UNIQUE_KEY_NAME:
            unique_features = [float(f) for f in row[1:]]

    unique_row = _UNIQUE_KEY_NAME in key_values
    assert len(key_values) > int(unique_row), 'There should be at least one row except {}'.format(_UNIQUE_KEY_NAME)

    key_counter = collections.Counter(key_values)
    most_common = key_counter.most_common(1)[0]
    assert most_common[1] == 1, 'Dublicate key found: {}'.format(most_common[0])

    if unique_row:
        key_values.remove(_UNIQUE_KEY_NAME)
    else:
        unique_features = [0.] * features_size
    key_values.insert(0, _UNIQUE_KEY_NAME)

    return key_values, features_size, unique_features


def read_and_split(file_name, unique_features, shard_rows):
    size, features = 1, unique_features

    for row in _iterate_text_vectors(file_name):
        if row[0] == _UNIQUE_KEY_NAME:
            continue

        if size == shard_rows:
            yield features
            size, features = 0, []

        features.extend([float(f) for f in row[1:]])
        size += 1

    if size > 0:
        yield features


def export_hub_module(src_path, dest, trainable=False, combiner='mean', max_norm=None, shard_bytes=2 ** 30):
    logging.info('Checking keys and vectors in {}'.format(src_path))
    key_values, embed_size, unique_features = read_and_count(src_path)
    logging.info('Done checking: rows_count={}, embeddings_size={}'.format(len(key_values), embed_size))

    if set(unique_features) == {0.}:
        logging.info('Embedding for {} initialized with zeros'.format(_UNIQUE_KEY_NAME))
    else:
        logging.info('Embedding for {} found, using provided values'.format(_UNIQUE_KEY_NAME))

    logging.info('Creating hub module specs')
    module_spec = _make_module_spec(key_values, embed_size, trainable, combiner, max_norm, shard_bytes)
    hub_module = hub.Module(module_spec)

    # The embeddings may be very large (e.g., larger than the 2GB serialized Tensor limit).
    # To avoid having them frozen as constant Tensors in the graph we instead assign them through the
    # placeholders and feed_dict mechanism.
    # Additionally we should use variable partitioning due to restoring issues in MacOS X.
    shard_rows = hub_module.variable_map[_EMBEDDINGS_VAR_NAME][0].shape[0]
    total_shards = len(hub_module.variable_map[_EMBEDDINGS_VAR_NAME])
    with tf.Session() as sess:
        for part_i, part_values in enumerate(read_and_split(src_path, unique_features, shard_rows)):
            logging.info('Initializing embeddings part: {} of {}'.format(part_i + 1, total_shards))

            init_values = tf.placeholder(dtype=tf.float32, shape=[shard_rows, embed_size])
            part_var = hub_module.variable_map[_EMBEDDINGS_VAR_NAME][part_i]
            load_values = tf.assign(part_var, init_values)

            part_values = np.array(part_values, dtype=np.float32).reshape((shard_rows, embed_size))
            sess.run([load_values], feed_dict={init_values: part_values})

        logging.info('Exporting hub module to {}'.format(dest))
        hub_module.export(dest, sess)


def _make_module_spec(keys_vocab, embedding_size, is_trainable, combiner, max_norm, shard_bytes):
    def module_fn():
        embeddings_shape = [len(keys_vocab), embedding_size]
        embedding_weights = tf.get_variable(
            name=_EMBEDDINGS_VAR_NAME,
            shape=embeddings_shape,
            dtype=tf.float32,
            initializer=tf.zeros_initializer(),
            trainable=is_trainable,
            partitioner=tf.variable_axis_size_partitioner(
                max_shard_bytes=shard_bytes  # 1Gb by default
            )
        )
        lookup_table = tf.contrib.lookup.index_table_from_tensor(mapping=keys_vocab, default_value=0)

        default_keys = tf.placeholder(dtype=tf.string, shape=[None])
        default_ids = lookup_table.lookup(default_keys)
        default_embeddings = tf.nn.embedding_lookup(
            params=embedding_weights,
            ids=default_ids,
            partition_strategy='div',
            max_norm=max_norm,
        )
        hub.add_signature('default', default_keys, default_embeddings)

        context_keys = tf.sparse_placeholder(dtype=tf.string, shape=(None, None))
        context_ids = lookup_table.lookup(context_keys)
        context_embeddings = tf.nn.safe_embedding_lookup_sparse(
            embedding_weights=embedding_weights,
            sparse_ids=context_ids,
            combiner=combiner,
            default_id=0,
            partition_strategy='div',
            max_norm=max_norm,
        )
        hub.add_signature('context', context_keys, context_embeddings)

        sequence_keys = tf.sparse_placeholder(dtype=tf.string, shape=(None, None, None))
        sequence_ids = lookup_table.lookup(sequence_keys)
        sequence_embeddings = tf.nn.safe_embedding_lookup_sparse(
            embedding_weights=embedding_weights,
            sparse_ids=sequence_ids,
            combiner=combiner,
            default_id=0,
            partition_strategy='div',
            max_norm=max_norm,
        )
        hub.add_signature('sequence', sequence_keys, sequence_embeddings)

    return hub.create_module_spec(module_fn)


def main():
    parser = argparse.ArgumentParser(
        description='Convert string keys and corresponding vectors of floats into TF Hub lookup module')
    parser.add_argument(
        'src_path',
        type=argparse.FileType('rb'),
        help='Path to saved NumPy or text 2D array. First column should contain string keys. '
             'First key should be "{}"'.format(_UNIQUE_KEY_NAME))
    parser.add_argument(
        'dest_path',
        type=str,
        help='Path to export TF Hub module')
    parser.add_argument(
        '--trainable',
        action='store_true',
        help='Whither embedding variable should be trainable')
    parser.add_argument(
        '--combiner',
        choices=['sum', 'mean', 'sqrtn'],
        default='mean',
        help='How to combine embedding results for each entry')
    parser.add_argument(
        '--max_norm',
        type=float,
        default=None,
        help='If not None, all embeddings are l2-normalized to max_norm before combining')

    argv, _ = parser.parse_known_args()

    src_path = argv.src_path.name
    argv.src_path.close()

    assert not os.path.exists(argv.dest_path) or os.path.isdir(argv.dest_path)
    try:
        os.makedirs(argv.dest_path)
    except:
        pass

    logging.basicConfig(level=logging.INFO)

    logging.info('Export options: trainable={}, combiner={}, max_norm={}'.format(
        argv.trainable, argv.combiner, argv.max_norm))
    export_hub_module(src_path, argv.dest_path, argv.trainable, argv.combiner, argv.max_norm)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from setuptools import setup, find_packages

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name='tfembedhub',
    version='1.2.1',
    description='TensorFlow Hub module producer for text embedding lookup',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/shkarupa-alex/tfembedhub',
    author='Shkarupa Alex',
    author_email='shkarupa.alex@gmail.com',
    license='MIT',
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'tfembedhub-convert=tfembedhub.export:main',
        ],
    },
    install_requires=[
        # https://github.com/tensorflow/tensorflow/issues/7166
        # 'tensorflow>=1.9.0',
        'tensorflow_hub>=0.1.1',
        'numpy>=1.13.3',
    ],
    extras_require={
        'tf_cpu': ['tensorflow>=1.9.0'],
        'tf_gpu': ['tensorflow-gpu>=1.9.0'],
    },
    test_suite='nose.collector',
    tests_require=['nose']
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import os
import shutil
import tempfile
import tensorflow as tf
import tensorflow_hub as hub
import unittest

from ..export import _UNIQUE_KEY_NAME, read_and_count, read_and_split, export_hub_module


class TestRead(unittest.TestCase):
    def setUp(self):
        self.temp_dir = tempfile.mkdtemp()
        self.temp_file = os.path.join(self.temp_dir, 'source.txt')

    def tearDown(self):
        shutil.rmtree(self.temp_dir, ignore_errors=True)

    def testReadAndCount(self):
        source = '\n'.join([
            'key1 1. 2.',
            'key2 3. 4.',
        ]) + '\n'
        with open(self.temp_file, 'wb') as ttf:
            ttf.write(source.encode('utf-8'))

        key_values, features_size, unique_features = read_and_count(self.temp_file)
        self.assertListEqual(key_values, [_UNIQUE_KEY_NAME, 'key1', 'key2'])
        self.assertEqual(features_size, 2)
        self.assertListEqual(unique_features, [0., 0.])

    def testReadAndSplit(self):
        source = '\n'.join([
            'key1 1. 2.',
            '{} -1. 0.'.format(_UNIQUE_KEY_NAME),
            'key2 3. 4.',
        ]) + '\n'
        with open(self.temp_file, 'wb') as ttf:
            ttf.write(source.encode('utf-8'))

        _, _, unique_features = read_and_count(self.temp_file)

        expected = [
            [-1.0, 0.0, 1.0, 2.0],
            [3.0, 4.0],
        ]
        actual = list(read_and_split(self.temp_file, unique_features, 2))

        self.assertEqual(actual, expected)


class TestExportHubModule(tf.test.TestCase):
    def setUp(self):
        self.temp_dir = tempfile.mkdtemp()
        self.temp_file = os.path.join(self.temp_dir, 'source.txt')
        tf.reset_default_graph()

    def tearDown(self):
        shutil.rmtree(self.temp_dir, ignore_errors=True)

    def testExport(self):
        source_keys = ['key{}'.format(i) for i in range(25)]
        source_values = np.array([i * 1. for i in range(50)]).reshape((25, 2)).tolist()

        source_content = '\n'.join(
            ' '.join([k] + [str(v) for v in values]) for k, values in zip(source_keys, source_values))
        with open(self.temp_file, 'wb') as ttf:
            ttf.write(source_content.encode('utf-8'))

        export_hub_module(self.temp_file, self.temp_dir, shard_bytes=16)
        lookup_values = hub.Module(self.temp_dir)

        expected_values_default = [[2., 3.]]
        lookup_values_default = lookup_values(['key1'])

        expected_values_context = [[2., 3.]]
        lookup_values_context = lookup_values(tf.SparseTensor(
            indices=[[0, 0]],
            values=['key1'],
            dense_shape=[1, 1]
        ), signature='context')

        expected_values_sequence = [[[2., 3.]]]
        lookup_values_sequence = lookup_values(tf.SparseTensor(
            indices=[[0, 0, 0]],
            values=['key1'],
            dense_shape=[1, 1, 1]
        ), signature='sequence')

        expected_values_all = [[0., 0.]] + source_values
        lookup_values_all = lookup_values([_UNIQUE_KEY_NAME] + source_keys)

        with self.test_session() as sess:
            sess.run([tf.global_variables_initializer(), tf.tables_initializer()])
            [
                result_values_default,
                result_values_context,
                result_values_sequence,
                result_values_all,
            ] = sess.run([
                lookup_values_default,
                lookup_values_context,
                lookup_values_sequence,
                lookup_values_all,
            ])

        self.assertListEqual(expected_values_default, result_values_default.tolist())
        self.assertListEqual(expected_values_context, result_values_context.tolist())
        self.assertListEqual(expected_values_sequence, result_values_sequence.tolist())
        self.assertListEqual(expected_values_all, result_values_all.tolist())
